

#include "BBCBillboardCloudUVMapper.h"
#include "BBCBillboardCloud.h"
#include "BBCEntityCluster.h"
#include "BBCBillboardClusterData.h"

namespace BBC {

BillboardCloudUVMapper::BillboardCloudUVMapper() 
{
}

BillboardCloudUVMapper::~BillboardCloudUVMapper()
{
	shutdown(); 
}

void BillboardCloudUVMapper::initialize()
{
}

void BillboardCloudUVMapper::shutdown() 
{
	while (getNumBsp() > 0)
	{
		removeBsp(getNumBsp()-1);
	}

	while (getNumClusterLists() > 0)
	{
		removeClusterList(getNumClusterLists()-1);
	}	
}

unsigned int BillboardCloudUVMapper::getNumClusterLists()
{
	return mClusterLists.size();
}

unsigned int BillboardCloudUVMapper::getNumBsp()
{
	return mBspList.size();
}

void BillboardCloudUVMapper::addBsp(IMG::Bsp *bsp)
{
	mBspList.push_back(bsp);
}

IMG::Bsp* BillboardCloudUVMapper::getBsp(unsigned int iBsp)
{
	return mBspList[iBsp];
}

void BillboardCloudUVMapper::removeBsp(unsigned int iBsp)
{
	IMG::Bsp *bsp = mBspList[iBsp];
	mBspList.erase(mBspList.begin() + iBsp);
	delete bsp;
}

void BillboardCloudUVMapper::addClusterList(IMG::ClusterList *clusterList)
{
	mClusterLists.push_back(clusterList);
}

IMG::ClusterList* BillboardCloudUVMapper::getClusterList(unsigned int iClusterList)
{
	return mClusterLists[iClusterList];
}

void BillboardCloudUVMapper::removeClusterList(unsigned int iClusterList)
{
	IMG::ClusterList *clusterList = mClusterLists[iClusterList];
	mClusterLists.erase(mClusterLists.begin() + iClusterList);
	delete clusterList;
}

void BillboardCloudUVMapper::setTextureWidth(unsigned short int width) 
{
	mTextureWidth = width;
}

unsigned short int BillboardCloudUVMapper::getTextureWidth() 
{
	return mTextureWidth;
}

void BillboardCloudUVMapper::setTextureHeight(unsigned short int height) 
{
	mTextureHeight = height;
}

unsigned short BillboardCloudUVMapper::getTextureHeight() 
{
	return mTextureHeight;
}

void BillboardCloudUVMapper::setTextureAtlasWidth(unsigned short int width) 
{
	mTextureAtlasWidth = width;
}

unsigned short int BillboardCloudUVMapper::getTextureAtlasWidth() 
{
	return mTextureAtlasWidth;
}

void BillboardCloudUVMapper::setTextureAtlasHeight(unsigned short int height) 
{
	mTextureAtlasHeight = height;
}

unsigned short BillboardCloudUVMapper::getTextureAtlasHeight() 
{
	return mTextureAtlasHeight;
}

void BillboardCloudUVMapper::addTextureAtlasTextureCoords(bool mergeBillboards, bool groupedBillboards, unsigned int numGeneratedBillboardClouds, unsigned int iBillboardGroup)
{
	unsigned int iUniqueVertex = 0;

	Ogre::LogManager::getSingleton().logMessage("iBBC:" + Ogre::StringConverter::toString(numGeneratedBillboardClouds) + "-" + "iBBCG:" + Ogre::StringConverter::toString(iBillboardGroup));

	Ogre::LogManager::getSingleton().logMessage("SizeCLList:" + Ogre::StringConverter::toString(getClusterList(iBillboardGroup)->getSize()) + " -- numBBC:" + Ogre::StringConverter::toString(numGeneratedBillboardClouds));

	IMG::Cluster *cluster = getClusterList(iBillboardGroup)->get(getClusterList(iBillboardGroup)->getSize()-1);		
	IMG::Box2d *box = cluster->getBspNodeBound();
	Ogre::Vector2 max = box->getMaximum();
	Ogre::Vector2 min = box->getMinimum();
	Ogre::LogManager::getSingleton().logMessage("Max:" + Ogre::StringConverter::toString(Ogre::Vector3(max[0],max[1],0)) + 
		"\tMin:" + Ogre::StringConverter::toString(Ogre::Vector3(min[0],min[1],0)) + 
		"\tSize(" + Ogre::StringConverter::toString(max[0]-min[0]) +
		"," + Ogre::StringConverter::toString(max[1]-min[1]) + ")");

	SubEntityPtr subEntity;												
	
	Ogre::Vector2 texCoords01,texCoords02,texCoords03,texCoords04;
	texCoords01[0] = min[0] / mTextureAtlasWidth;
	texCoords01[1] = min[1] / mTextureAtlasHeight;

	texCoords02[0] = max[0] / mTextureAtlasWidth;
	texCoords02[1] = min[1] / mTextureAtlasHeight;

	texCoords03[0] = max[0] / mTextureAtlasWidth;
	texCoords03[1] = max[1] / mTextureAtlasHeight;

	texCoords04[0] = min[0] / mTextureAtlasWidth;
	texCoords04[1] = max[1] / mTextureAtlasHeight;
	
	if (groupedBillboards)
	{
		Ogre::LogManager::getSingleton().logMessage("DontMerge");
		subEntity = mBillboardCloud->getEntity()->getSubEntity(iBillboardGroup);				
		
		// Face A
		//(0,0,0);
		//(1,0,0);
		//(1,1,0);
		subEntity->setTexCoord(iUniqueVertex, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords01[0],texCoords01[1],0));
		subEntity->setTexCoord(iUniqueVertex+1, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords02[0],texCoords02[1],0));
		subEntity->setTexCoord(iUniqueVertex+2, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords03[0],texCoords03[1],0));

		Ogre::ColourValue subEntityColour01 = Ogre::ColourValue(0, texCoords01[0],texCoords01[1]);
		Ogre::ColourValue subEntityColour02 = Ogre::ColourValue(0, texCoords02[0],texCoords02[1]);
		Ogre::ColourValue subEntityColour03 = Ogre::ColourValue(0, texCoords03[0],texCoords03[1]);

		subEntity->setVertexColour(iUniqueVertex, subEntityColour01.getAsRGBA());
		subEntity->setVertexColour(iUniqueVertex+1, subEntityColour02.getAsRGBA());
		subEntity->setVertexColour(iUniqueVertex+2, subEntityColour03.getAsRGBA());

		// Face B
		//(1,1,0);
		//(0,1,0);
		//(0,0,0);
		subEntity->setTexCoord(iUniqueVertex+3, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords03[0],texCoords03[1],0));
		subEntity->setTexCoord(iUniqueVertex+4, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords04[0],texCoords04[1],0));
		subEntity->setTexCoord(iUniqueVertex+5, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords01[0],texCoords01[1],0));

		Ogre::ColourValue subEntityColour04 = Ogre::ColourValue(0, texCoords03[0],texCoords03[1]);
		Ogre::ColourValue subEntityColour05 = Ogre::ColourValue(0, texCoords04[0],texCoords04[1]);
		Ogre::ColourValue subEntityColour06 = Ogre::ColourValue(0, texCoords01[0],texCoords01[1]);

		subEntity->setVertexColour(iUniqueVertex+3, subEntityColour04.getAsRGBA());
		subEntity->setVertexColour(iUniqueVertex+4, subEntityColour05.getAsRGBA());
		subEntity->setVertexColour(iUniqueVertex+5, subEntityColour06.getAsRGBA());

		iUniqueVertex = iUniqueVertex + 6;
	}
	else
	{
		if (!mergeBillboards)
		{
			Ogre::LogManager::getSingleton().logMessage("DontMerge");

			subEntity = mBillboardCloud->getEntity()->getSubEntity(numGeneratedBillboardClouds);
			subEntity->addTextureCoordSet(2);
			subEntity->enableVertexColours(true);
			// Face A
			//(0,0,0);
			//(1,0,0);
			//(1,1,0);

			subEntity->setTexCoord(0, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords01[0],texCoords01[1], 0));
			subEntity->setTexCoord(1, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords02[0],texCoords02[1], 0));
			subEntity->setTexCoord(2, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords03[0],texCoords03[1], 0));

			Ogre::ColourValue subEntityColour01 = Ogre::ColourValue(0, texCoords01[0],texCoords01[1]);
			Ogre::ColourValue subEntityColour02 = Ogre::ColourValue(0, texCoords02[0],texCoords02[1]);
			Ogre::ColourValue subEntityColour03 = Ogre::ColourValue(0, texCoords03[0],texCoords03[1]);

			subEntity->setVertexColour(0, subEntityColour01.getAsRGBA());
			subEntity->setVertexColour(1, subEntityColour02.getAsRGBA());
			subEntity->setVertexColour(2, subEntityColour03.getAsRGBA());

			// Face B
			//(1,1,0);
			//(0,1,0);
			//(0,0,0);
			subEntity->setTexCoord(3, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords03[0],texCoords03[1],0));
			subEntity->setTexCoord(4, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords04[0],texCoords04[1],0));
			subEntity->setTexCoord(5, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords01[0],texCoords01[1],0));

			Ogre::ColourValue subEntityColour04 = Ogre::ColourValue(0, texCoords03[0],texCoords03[1]);
			Ogre::ColourValue subEntityColour05 = Ogre::ColourValue(0, texCoords04[0],texCoords04[1]);
			Ogre::ColourValue subEntityColour06 = Ogre::ColourValue(0, texCoords01[0],texCoords01[1]);

			subEntity->setVertexColour(3, subEntityColour04.getAsRGBA());
			subEntity->setVertexColour(4, subEntityColour05.getAsRGBA());
			subEntity->setVertexColour(5, subEntityColour06.getAsRGBA());

		}
		else
		{
			Ogre::LogManager::getSingleton().logMessage("Merge");

			subEntity = mBillboardCloud->getEntity()->getSubEntity(0);				
			
			// Face A
			//(0,0,0);
			//(1,0,0);
			//(1,1,0);
			subEntity->setTexCoord(iUniqueVertex, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords01[0],texCoords01[1],0));
			subEntity->setTexCoord(iUniqueVertex+1, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords02[0],texCoords02[1],0));
			subEntity->setTexCoord(iUniqueVertex+2, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords03[0],texCoords03[1],0));

			Ogre::ColourValue subEntityColour01 = Ogre::ColourValue(0, texCoords01[0],texCoords01[1]);
			Ogre::ColourValue subEntityColour02 = Ogre::ColourValue(0, texCoords02[0],texCoords02[1]);
			Ogre::ColourValue subEntityColour03 = Ogre::ColourValue(0, texCoords03[0],texCoords03[1]);

			subEntity->setVertexColour(iUniqueVertex, subEntityColour01.getAsRGBA());
			subEntity->setVertexColour(iUniqueVertex+1, subEntityColour02.getAsRGBA());
			subEntity->setVertexColour(iUniqueVertex+2, subEntityColour03.getAsRGBA());

			// Face B
			//(1,1,0);
			//(0,1,0);
			//(0,0,0);
			subEntity->setTexCoord(iUniqueVertex+3, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords03[0],texCoords03[1],0));
			subEntity->setTexCoord(iUniqueVertex+4, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords04[0],texCoords04[1],0));
			subEntity->setTexCoord(iUniqueVertex+5, subEntity->getNumTexCoordSets()-1, Ogre::Vector3(texCoords01[0],texCoords01[1],0));

			Ogre::ColourValue subEntityColour04 = Ogre::ColourValue(0, texCoords03[0],texCoords03[1]);
			Ogre::ColourValue subEntityColour05 = Ogre::ColourValue(0, texCoords04[0],texCoords04[1]);
			Ogre::ColourValue subEntityColour06 = Ogre::ColourValue(0, texCoords01[0],texCoords01[1]);

			subEntity->setVertexColour(iUniqueVertex+3, subEntityColour04.getAsRGBA());
			subEntity->setVertexColour(iUniqueVertex+4, subEntityColour05.getAsRGBA());
			subEntity->setVertexColour(iUniqueVertex+5, subEntityColour06.getAsRGBA());

			iUniqueVertex = iUniqueVertex + 6;
		}
		Ogre::LogManager::getSingleton().logMessage("EndAdd");
	}
}

void BillboardCloudUVMapper::addTextureInBsp(unsigned int iBillboardGroup, unsigned int iBillboard)
{
	IMG::NodeBspPtr nodeDest;
	IMG::Cluster cluster;
	Ogre::LogManager::getSingleton().logMessage("w:" + Ogre::StringConverter::toString(mTextureWidth));
	Ogre::LogManager::getSingleton().logMessage("h:" + Ogre::StringConverter::toString(mTextureHeight));
	nodeDest = getBsp(iBillboardGroup)->insert(mTextureWidth,mTextureHeight,iBillboard);
	
	cluster.setBspNodeBound(nodeDest->getBound());
	//cluster.finish();
	getClusterList(iBillboardGroup)->insert(&cluster);	
}

void BillboardCloudUVMapper::generateTextureAtlasTextureCoords(bool mergeBillboards, bool groupedBillboards, unsigned int numGroups) 
{	
	unsigned int numBillboardGroups;
	bool enabledTextureCoordSet = false;
	unsigned int numGeneratedBillboardClouds = 0;

	for (unsigned int iBillboard = 0; iBillboard < mBillboardCloud->getNumBillboards(); iBillboard++)
	{
		BillboardPtr billboard = mBillboardCloud->getBillboard(iBillboard);		
		EntityClusterPtr entityCluster = billboard->getBillboardClusterData()->getEntityCluster();

		if (entityCluster->getNumEntitiesClusterData() > 0)
		{
			if (mergeBillboards)
			{
				if (!enabledTextureCoordSet)
				{
					SubEntityPtr subEntity = mBillboardCloud->getEntity()->getSubEntity(0);	
					subEntity->addTextureCoordSet(2);
					subEntity->enableVertexColours(true);
					enabledTextureCoordSet = true;
				}
			}
			else
			{
				if (groupedBillboards)
				{
					SubEntityPtr subEntity = mBillboardCloud->getEntity()->getSubEntity(numGeneratedBillboardClouds);	
					subEntity->addTextureCoordSet(2);
					subEntity->enableVertexColours(true);
				}
			}

			numGeneratedBillboardClouds++;
		}
	}

	if (groupedBillboards)
	{
		numBillboardGroups = numGroups;
	}
	else
	{
		if (mergeBillboards)
		{
			numBillboardGroups = 1;
		}
		else // (!mergeBillboards)
		{
			numBillboardGroups = 1;
		}
	}
	
	unsigned int iBillboardsInGroup = 0;
	unsigned int numberBillboardsEachGroup = numGeneratedBillboardClouds / numGroups;
	unsigned int numberBillboardsAddedFirstGroup = numGeneratedBillboardClouds % numGroups;
	numGeneratedBillboardClouds = 0;
	unsigned int iBillboardGroup;
	for (iBillboardGroup = 0; iBillboardGroup < numBillboardGroups; iBillboardGroup++)
	{		
		IMG::Bsp *bsp = new IMG::Bsp();
		IMG::ClusterList *clusterList = new IMG::ClusterList();
		IMG::NodeBspPtr node = IMG::NodeBspPtr(new IMG::NodeBsp);

		node->getBound()->setBoundBox(0, 0, mTextureAtlasWidth, mTextureAtlasHeight);
		bsp->setRoot(node);	

		addBsp(bsp);
		addClusterList(clusterList);

	}

	iBillboardGroup = 0;
	for (unsigned int iBillboard = 0; iBillboard < mBillboardCloud->getNumBillboards(); iBillboard++)
	{
		Ogre::LogManager::getSingleton().logMessage("id:" + Ogre::StringConverter::toString(mBillboardCloud->getBillboard(iBillboard)->getBillboardHandle()));

		BillboardClusterData *billboardClusterData = mBillboardCloud->getBillboard(iBillboard)->getBillboardClusterData().get();		
		
		if (billboardClusterData != NULL)
		{
			Ogre::LogManager::getSingleton().logMessage("billboardData");
			EntityClusterPtr entityCluster = billboardClusterData->getEntityCluster();
			if (entityCluster->getNumEntitiesClusterData() > 0)
			{			
				Ogre::LogManager::getSingleton().logMessage("entityCluster");
				
				if ((iBillboardGroup == 0) && (iBillboardsInGroup < (numberBillboardsEachGroup + numberBillboardsAddedFirstGroup)))
				{
					Ogre::LogManager::getSingleton().logMessage("\nA");
					Ogre::LogManager::getSingleton().logMessage("IG:" + Ogre::StringConverter::toString(iBillboardsInGroup) + " SizeCLList(" + Ogre::StringConverter::toString(iBillboardGroup) + "):" + Ogre::StringConverter::toString(getClusterList(iBillboardGroup)->getSize()) + " -- numBBC:" + Ogre::StringConverter::toString(numGeneratedBillboardClouds));

					addTextureInBsp(iBillboardGroup, iBillboard);

					addTextureAtlasTextureCoords(mergeBillboards, groupedBillboards, numGeneratedBillboardClouds, iBillboardGroup);
					iBillboardsInGroup++;
				}
				else
				{
					if ((iBillboardGroup > 0) && (iBillboardsInGroup < numberBillboardsEachGroup))
					{
						Ogre::LogManager::getSingleton().logMessage("\nB");
						Ogre::LogManager::getSingleton().logMessage("IG:" + Ogre::StringConverter::toString(iBillboardsInGroup) + " SizeCLList(" + Ogre::StringConverter::toString(iBillboardGroup) + "):" + Ogre::StringConverter::toString(getClusterList(iBillboardGroup)->getSize()) + " -- numBBC:" + Ogre::StringConverter::toString(numGeneratedBillboardClouds));

						addTextureInBsp(iBillboardGroup, iBillboard);

						addTextureAtlasTextureCoords(mergeBillboards, groupedBillboards, numGeneratedBillboardClouds, iBillboardGroup);
						iBillboardsInGroup++;
					}
					else
					{
						iBillboardGroup++;
						iBillboardsInGroup = 0;

						Ogre::LogManager::getSingleton().logMessage("\nC IG:" + Ogre::StringConverter::toString(iBillboardsInGroup-1) + " SizeCLList(" + Ogre::StringConverter::toString(iBillboardGroup-1) + "):" + Ogre::StringConverter::toString(getClusterList(iBillboardGroup-1)->getSize()));
						Ogre::LogManager::getSingleton().logMessage("IG:" + Ogre::StringConverter::toString(iBillboardsInGroup) + " SizeCLList(" + Ogre::StringConverter::toString(iBillboardGroup) + "):" + Ogre::StringConverter::toString(getClusterList(iBillboardGroup)->getSize()) + " -- numBBC:" + Ogre::StringConverter::toString(numGeneratedBillboardClouds));

						addTextureInBsp(iBillboardGroup, iBillboard);

						addTextureAtlasTextureCoords(mergeBillboards, groupedBillboards, numGeneratedBillboardClouds, iBillboardGroup);
						iBillboardsInGroup++;
					}
				}

				numGeneratedBillboardClouds++;
			}
		}
	}	
	//mBsp.Print();
}

void BillboardCloudUVMapper::setBillboardCloud(BillboardCloud *billboardCloud) 
{
	mBillboardCloud = billboardCloud;
}

BillboardCloud * BillboardCloudUVMapper::getBillboardCloud() 
{
	return mBillboardCloud;
}

}
